
;; ------------------------------------------------------------
;; William R. Bevier and Michael K. Smith.
;; Copyright 1995, 1996 Computational Logic, Inc. All Rights Reserved
;; ------------------------------------------------------------

(require "general")
(require "conditions-events-actions")
(require "system")
(require "nodes")
(require "arcs")
(provide "step")

(defvar *interactive* nil)

(defstruct system-delta
  events alist)

;; ------------------------------------------------------------
;; Notes
;
;;  We assume that all arcs in the model take a non-empty
;;  set of state nodes (ie. nodes of type AND, OR, or BASIC) to a non-empty 
;;  set of state nodes. 
;
;;  Do we care about contradictory events in the current events list?
;;  Do we care if evaluation of the environmental action in the first
;;    step overwrites previous conditions and events?
;
;;  We assumed that it would be ok to have a single entry and exit
;;  action. If you wanted to do multiple actions on entry or exit,
;;  wrap a PAR around them. 
;
;; We exclude termination nodes and connectors.
;
;; We have not dealt with time.
;; ------------------------------------------------------------
;

;; ------------------------------------------------------------
;; RUN-SYSTEM:  simulator
;; STEP-SYSTEM: semantics of a step.
;; From Semantics of Statecharts, Jan 91, pg 26.
;; ------------------------------------------------------------

(defun run-system (system)
  (run-system1 (initial-system-step system)))

(defun run-system1 (system)
  (cond ((system-halted system) system)
	(t (run-system1 (step-system system)))))

(defvar *max-time* 100000)

(defvar *halt-system* 'halt)

(defun system-halted (s)
  (let ((time (system-time s))
	(schedule (system-schedule s)))
    (cond ((> time *max-time*)
	   (status-report "~%Out of time.") t)
	  ((and schedule
		(equal (car (car schedule)) time)
		(equal (cdr (car schedule)) *halt-system*))
	   (status-report "~%Encountered halt event.") t)
	  ((null (system-configuration s))
	   (status-report "~%Empty configuration.") t))))

(defvar *nodes* nil "Set by initial-system-step")
;; (defvar *arcs*)

(defun initial-system-step (system)
  ;; The system is in an initial configuration.  
  ;; Make sure it is consistent.
  (let* ((conf (system-configuration system))
	 (time (system-time system))
	 (nodes (initialize-node-ticks (system-nodes system)))
	 (new-config (union conf (enabled-nodes conf nodes)))
         (deltas     (iterate for nm in new-config
			      collect (do-enter-action nm nodes system))))
    (setq *time* time)
    (setq *nodes* nodes)
    (mapcar (f-l (x)
		 (let ((node (find-node x (system-nodes system))))
		   (setf (node-busy node) 0)
		   (setf (node-idle node) time)))
	    new-config)
    (setf (system-configuration system) new-config)
    (setq system (commit-delta-to-system (accum-system-deltas* deltas)
					 system))
    (if (configurationp new-config (system-nodes system))  
	system
	(error (format nil "Inconsistent inital SYSTEM configuration ~s" conf)))))

(defun display-enabled (enabled) (display-enabled-arc (car enabled)))

(defun step-system (system)
  (if (system-halted system) 
      (progn (format nil "System encountered halt event.") system)
      (let ((enabled (enabled-arcs system)))
	;; Enabled is list whose elements are (arc . bindings)
	(if *interactive*
	    (if enabled
		(mapcar #'display-enabled enabled)
		(status-report "~%No arcs enabled.")))
	(execute-step (remove-conflicting-enabled enabled system) system))))

;; ------------------------------------------------------------
;; Support functions
;; ------------------------------------------------------------

(defun current-events (schedule time)
  (iterate for ev in schedule 
	   until (> (car ev) time)
	   collect (cdr ev)))

(defun remove-current-events (system time)
  (setf (system-schedule system) (deschedule (system-schedule system) time))
  system)

(defvar *event-history* nil
  "Very inefficient implementation for TIMEOUT mechanism. Fix
at later date.  List of (event . time).")

(defun deschedule (schedule time)
  (cond ((null schedule) nil)
	((<= (car (car schedule)) time)
	 (setq *event-history* (cons (cons (cdar schedule) time) *event-history*))
	 (deschedule (cdr schedule) time))
	(t schedule)))

(defun lessp-car (a b) (< (car a) (car b)))

(defun schedule-immediate (new sched)
  (sort (append new sched) #'lessp-car))

(defun schedule-events (new sched)
  (cond ((null new) sched)
	((null sched) new)
	((< (car (car new)) (car (car sched)))
	 (cons (car new) (schedule-events (cdr new) sched)))
	(t (cons (car sched) (schedule-events new (cdr sched))))))

(defun accum-system-deltas (new old)
  (make-system-delta :events (union (system-delta-events new) (system-delta-events old)
				    :test #'equal)
		     :alist (merge-alists (system-delta-alist new)
					     (system-delta-alist old)
					     t)))
				    
(defun accum-system-deltas* (l)
  (cond ((atom l) nil)
	((atom (cdr l)) (car l))
	(t (accum-system-deltas (car l) (accum-system-deltas* (cdr l))))))

(defun arc-priority (arc nodes)
  (nlevel (arc-scope arc)
	  nodes))

(defun commit-delta-to-system (delta system)
  ;; Commit delta to activities to system
  ;; (status-report "~%Delta ~a" delta)
  (if delta
      (update-system system
		     :schedule (schedule-events (system-delta-events delta)
						(system-schedule system))
		     :alist (bind-all (system-delta-alist delta)
				      (system-alist system)))
      system))

;; CONFLICTING-ARCSP
;; Two arcs conflict when both cause the system to exit
;; a common state. The Semantics of Statecharts. Pg. 24.

(defun conflicting-arcsp (a1 a2 nodes)
  (intersection (exiting-states a1 nodes)
		(exiting-states a2 nodes)))

(defun conflicting-subsets (arcs system)
  (cond ((atom arcs) nil)
	(t (let ((l (cons (car arcs)
			  (find-conflicting-arcs (car arcs)
						   (cdr arcs)
						   (system-nodes system)))))
	     (cons l
		   (conflicting-subsets 
		    (set-difference arcs l :test #'equal)
		    system))))))

(defun debug-print-system (string system)
  (status-report "~%~a ~%  Events: ~a~%  Conditions: ~a~%  Vars: ~a" 
	  string (system-schedule system) (system-conditions system) (system-alist system)))

(defun debug-print-delta (string delta)
  (if (stringp string)
      (let ((ev (system-delta-events delta))
	    (va (system-delta-alist delta)))
	(if (or ev va)
	    (status-report "~%~a ~%  New events: ~a~%  New bindings: ~a" string ev va)
	    (status-report "~%~a" string)))))

(defun enabled-arcs (system)
  (let ((config (system-configuration system))
	(events (current-events (system-schedule system) (system-time system)))
	(alist (system-alist system))
	(time (system-time system))
	enabled)
    (if events (status-report "~%  Current Events ~s" events))
    (iterate for arc in (system-arcs system)
	     when (and (member (arc-from arc) config)
		       (or (null (arc-trigger arc))
			   ;; Enabled is either nil or a list of form
			   ;; (t . bindings)
			   (setq enabled (system-eval-event (arc-trigger arc)
							    events time config alist))))
	     collect (cons arc (cdr enabled)))))

;; EXECUTE-STEP
;;
;; Note the ordering on actions (in case we allow conflicts).
;; All actions are executed in the state of the initial SYSTEM.
;;
;;  Computed the entered and exited nodes.
;;  For each exited node,  execute its exiting action.
;;  For each arc execute the arc action
;;  For each entered node, execute its entering action.
;;  Compute the new configuration.
;;  Commit all the accumulated changes, and set the new configuration
;;

(defvar *time* 0)

(defun execute-step (enabled system)
  (let* ((time  (system-time system))
	 (nodes (system-nodes system))
	 (arcs  (mapcar #'car enabled))
	 (exit  (exited-nodes arcs nodes (system-configuration system)))
	 (enter (entered-nodes arcs nodes))
	 ;; Deltas are of the form 
	 (exit-deltas
	  (iterate for nm in exit
		   collect (do-exit-action nm nodes system)))
	 (arc-deltas
	  (iterate for enable in enabled
		   collect (do-eval-action enable system)))
	 (enter-deltas
	  (iterate for nm in enter
		   collect (do-enter-action nm nodes system)))
	 (new-config
	  (enter-exit-configuration (system-configuration system)
				    (full-configuration exit nodes)
				    (full-configuration enter nodes))))
    (setq *time* time)
    (update-node-ticks enter exit nodes time)
    (update-system (commit-delta-to-system
		    (accum-system-deltas* (append exit-deltas
						  arc-deltas
						  enter-deltas))
		    (remove-current-events system time))
		   :configuration new-config
		   :time (1+ time))))

(defun enter-exit-configuration (existing exit enter)
  (cond ((and (null exit) (null enter)) existing)
	((null exit) (union existing enter))
	((null enter) (set-difference existing exit))
	(t (union (set-difference existing exit) enter))))
    
(defun exists-higher-priority-arc (priority arcs nodes)
  (iterate for a in arcs
    thereis (< (arc-priority a nodes) priority)))

(defun enabled-nodes (node-names nodes)
  (bags-to-set
   (mapcar (f-l (nm) (arc-activate-nodes nm nil nodes))
	   node-names)))

(defun exited-nodes (arcs nodes config)
  (bags-to-set
   (mapcar (f-l (a) (arc-deactivate-nodes (arc-from a) (arc-scope a) nodes config))
	   arcs)))

(defun entered-nodes (arcs nodes)
  (bags-to-set
   (mapcar (f-l (a) (arc-activate-nodes (arc-to a) (arc-scope a) nodes))
	   arcs)))


;; Scope is deepest containing node of arc.

(defun arc-deactivate-nodes (from scope nodes config)
  (append (deactivate-down from nodes config)
	  (deactivate-up (nparent from nodes) scope nodes)))

(defun arc-activate-nodes (to scope nodes)
  (append (activate-down to nodes)
	  (activate-up (nparent to nodes) scope nodes)))

(defun deactivate-down (from nodes config)
  (cons from 
	(iterate for nm in (intersection (nchildren from nodes) config)
		 append (deactivate-down nm nodes config))))

(defun deactivate-up (nm scope nodes)
  (if (equal nm scope)
      nil
      (let ((n (find-node nm nodes)))
	(cons nm
	      (append (if (equal (node-type n) 'and)
			  (node-children n)
			  nil)
		      (deactivate-up (node-parent n) scope nodes))))))

(defun activate-down (to nodes)
  (let* ((n (find-node to nodes))
	 (init (node-initial n))
	 (ands (if (equal (node-type n) 'and) (node-children n) nil))
	 (subs (if init (cons init ands) ands)))
    (cons to 
	  (iterate for nm in subs
		   append (activate-down nm nodes)))))

(defun activate-up (nm scope nodes)
  (if (equal nm scope)
      nil
      (cons nm
	    (activate-up (node-parent (find-node nm nodes)) scope nodes))))

(defvar *bindings* nil "Used to bind (? vars) in events for later use by actions.")

(defun do-exit-action (nm nodes system) 
  (let* ((nd (find-node nm nodes))
	 *bindings*)
    (eval-wrapper (system-eval-action (node-exit nd) system)
		  :debug (format nil "Exit action for node ~a." nm)
		  :step  (format nil "Exiting  node ~a. [i/b : ~d/~d]."
				 nm (compute-node-idle nd *time*)(node-busy nd))
		  :debug-fn #'debug-print-delta
		  :step-fn #'debug-print-delta)))

(defun do-enter-action (nm nodes system)
  (let* ((nd (find-node nm nodes))
	 *bindings*)
    (eval-wrapper (system-eval-action (node-enter nd) system)
		  :debug (format nil "Enter action for node ~a." nm)
		  :step  (format nil "Entering node ~a. [b/i : ~d/~d]."
				 nm (node-busy nd)(compute-node-idle nd *time*))
		  :debug-fn #'debug-print-delta
		  :step-fn #'debug-print-delta)))

(defun do-eval-action (enable system)
  (let* ((arc (car enable))
	 (*bindings* (cdr enable))
	 (action (arc-action arc)))
    (eval-wrapper (system-eval-action action system)
		  :debug (if action (format nil "Arc action for arc ~a." (Arc-name arc)) nil)
		  :step  (if action (format nil "Action for arc ~a." (Arc-name arc)) nil)
		  :debug-fn #'debug-print-delta
		  :step-fn  #'debug-print-delta)))

(defun update-node-ticks (enter exit nodes time)
  ;; Time is system-time in ticks.
  (let (;; (really-entered (set-difference enter exit :test #'equal))
	(really-exited (set-difference exit enter :test #'equal)))
    ;; (iterate for nm in really-entered
    ;;   do (let ((nd (find-node nm nodes)))
    ;;		  (setf (node-busy nd) time)))
    ;; 		  (setf (node-idle nd) time)))
    (iterate for nm in enter
	     do (let ((nd (find-node nm nodes)))
		  (setf (node-idle nd) time)
		  (setf (node-busy nd) (+ 1 (node-busy nd)))))
    (iterate for nm in really-exited
	     do (let ((nd (find-node nm nodes)))
		  (setf (node-idle nd) -1)
		  (setf (node-busy nd) -1)))))

(defun initialize-node-ticks (nodes)
  ;; Time is 0, system-time in ticks.
  (iterate for nd in nodes
	   do (progn
		(setf (node-idle nd) 0)
		(setf (node-busy nd) -1)))
  nodes)

;; EXITING-STATES: return the set of states exited by an arc.

(defun exiting-states (arc nodes)
  (union (list (arc-from arc))
	 (union (set-difference (nancestors (arc-from arc) nodes)
				(nancestors (arc-to arc) nodes))
		(ndescendents (arc-from arc) nodes))))

;; FIND-CONFLICTING-ARCS
;; Finds all elements in ARCS that conflict with ARC.

(defun find-conflicting-arcs (arc arcs nodes)
  (iterate for a in arcs
    when (and (not (eq (arc-name a) (arc-name arc)))
	      (conflicting-arcsp arc a nodes))
    collect a))

(defun maximal-nonconflicting-arcs (arcs system)
  (list-product (conflicting-subsets arcs system)))

(defun merge-delta-into-system (delta system)
  ;; Merge delta into system.
  ;; Old values that are not overwritten persist.
  (update-system system
	      :schedule (union (system-delta-events delta)
			       (system-schedule system)
			       :test #'equal)
	      :alist (bind-all (system-delta-alist delta)
			       (system-alist system))))

(defun new-configuration (arc-set system)
  (let ((en (unionx (iterate for arc in arc-set
		      collect (arc-to arc))))
	(ex (unionx (iterate for arc in arc-set
		      collect (arc-from arc))))
	(nodes (system-nodes system)))
    (update-system system
		:configuration
		(enter-exit-configuration (system-configuration system)
					  (full-configuration en nodes)
					  (full-configuration ex nodes)))))

(defun remove-conflicting-enabled (enabled system)
  (let ((enabled (remove-lower-enabled enabled system)))
    (select-conflicting-equal-enabled enabled system)))

(defun remove-lower-enabled (enabled system)
  (let ((arcs (mapcar #'car enabled)))
    (iterate for e in enabled
	     when (let ((a (car e)))
		    (not (exists-higher-priority-arc
			  (arc-priority a (system-nodes system))
			  (find-conflicting-arcs a arcs (system-nodes system))
			  (system-nodes system))))
	     collect e)))

(defun select-conflicting-equal-enabled (enabled system)
  (let ((arcs (mapcar #'car enabled))
	selected)
    (iterate for e in enabled
	     when (let* ((a (car e))
			 (b (find-conflicting-arcs a arcs (system-nodes system)))
			 c)
		    (if (or (member a selected) (null b))
			t
			(progn (setq c (select-from-conflict a b))
			       (setq selected (cons c selected))
			       (equal c a))))
	     collect e)))

(defun select-from-conflict (a b)
  (cond ((null b) a)
	(*interactive* (select-by-number "The first arc conflicts with the others.  Choose one."
					 #'display-arc (cons a b)))
	(t (car (last b)))))

(defvar *eval-mode* nil)

(defun select-step (arc-sets system)
  ;; Return a set of arcs.
  (declare (ignore system))
  (case (length arc-sets)
    (0 nil)
    (1 (car arc-sets))
    (t (status-report "~%Nondeterminism Encountered")
       (if (equal *eval-mode* 'debug)
	   (mapc (function
		  (lambda (x)
		    (status-report "~%   Candidate arcs: ~{~a ~}:" (mapcar #'arc-name x))))
		 arc-sets))
       (if (equal *eval-mode* 'interactive)
	   (progn
	     (iterate for arc-set in arc-sets
		      as i from 0
		      do (status-report "~%  ~d Candidate arcs: ~{~a ~}:"
				 i (mapcar #'arc-name arc-set)))
	     (let ((choice (read)))
	       (cond ((and (numberp choice)
			   (< choice (length arc-sets)))
		      (nth choice arc-sets))
		     (t (status-report "Choice out of range. Picking first transition.")
			(car arc-sets)))))
	 (car arc-sets)))))

(defun system-eval-action (form system)
  (if (null form)
      (make-system-delta)
      (multiple-value-bind (evs globals)
	  (eval-action form system)
	(make-system-delta :events evs :alist globals))))

(defun system-eval-event (form events time conf alist)
  (cond ((null form) '(t))
	(t (eval-event form events time conf alist))))

